import re
from toolz.curried import pipe, curry
from datetime import date
from typing import Callable

__all__ = (
    "today_str resub decamel invalid_first_char replace_non_word "
    "conseq_char trim fmt_str".split()
)
def today_str(pattern: str="%Y-%m-%d"):
    """Formats Today's Date as a string given a strftime pattern."""
    return date.today().strftime(pattern)

@curry
def resub(pat: str, repl: str, string: str):
    """curried version of re.sub. Used to build up several string
    replacements into a single function. See fmt_str.
    """
    return re.sub(pat, repl, string)

@curry
def decamel(string: str, replace_char: str="_") -> str:
    """Adds a sep character to a camelCase string.
    >>> decamel('FirstNameMD')
    'First_Name_MD'
    """
    return resub("(?<=[a-z])(?P<match>[A-Z])",
        f"{replace_char}\g<match>", string)

@curry
def invalid_first_char(string: str, append_char: str="x") -> str:
    """Appends an valid character invalid characters at the start of a string.
    >>> invalid_first_char('1stLien')
    'x1stLien'
    """
    return resub("(?P<match>^[0-9])", f"{append_char}\g<match>", string)

@curry
def replace_non_word(string: str, replace_char:str="_") -> str:
    """Replaces any character that is not a-z or A-Z or 0-9 or _
    with the replace_char.
    >>> replace_non_word('Over/Under')
    'Over_Under'
    >>> replace_non_word('% Total')
    '__Total'
    """
    return resub("\W", f"{replace_char}", string)

@curry
def conseq_char(string: str, chars: str="_ ") -> str:
    """Replaces consecutive characters with a single character.
    >>> conseq_char('First__Name___MD')
    'First_Name_MD'
    """
    result = string
    for char in chars:
        result = resub(f"{char}+", f"{char}", result)
    return result

@curry
def trim(string: str, char: str="[\s_]") -> str:
    """Replaces characters at the beginning and end of string.
    By default will replace white space and "_" characters.
    >>> trim('_First_Name_')
    'First_Name'
    >>> trim(' first name   ')
    'first name'
    """
    return resub(f"^{char}+|{char}+$", "", string)

# If this is changed, update the doc string for fmt_str
FMT_STR_COLUMN_FORMATS = (decamel, invalid_first_char, replace_non_word,
    conseq_char, trim)

FMT_STR_PRETTY_COLUMN_FORMATS = (decamel, invalid_first_char, replace_non_word,
    conseq_char, trim, resub("_", " "))

FMT_STR_ATTR_FORMATS = (decamel, invalid_first_char, replace_non_word,
    conseq_char)

StrFunc = Callable[[str],str]

def fmt_str(string, *funcs: StrFunc) -> str:
    """Applies a series of functions to string. By default, uses the
    FMT_STR_COLUMN_FORMATS tuple which contains the following functions which are
    applied in order:
    1. decamel
    2. invalid_first_char
    3. replace_non_word
    4. conseq_char
    5. trim

    tests = ["  1stLien", "% Loans  ", "_Over/Under", "First Name", "FirstNameMD"]...
    >>> fmt_str('  1stLien')
    '1st_Lien'
    >>> fmt_str('% Loans  ')
    'Loans'
    >>> fmt_str('_Over/Under')
    'Over_Under'
    >>> fmt_str('First Name')
    'First_Name'
    >>> fmt_str('FirstNameMD')
    'First_Name_MD'

    User can provide own list of funcs instead of the defaults.
    However, each func must accept only 1 arg. funcs are applied from
    left to right.

    >>> fmt_str('FirstName/Company', decamel, replace_non_word, str.lower)
    'first_name_company'
    """
    if len(funcs) == 0:
        return pipe(string, *FMT_STR_COLUMN_FORMATS)
    else:
        return pipe(string, *funcs)


# AUTOGENERATED! DO NOT EDIT! File to edit: xlsx.ipynb (unless otherwise specified).

__all__ = ['to_xlsx', 'column_format', 'default_formats', 'column_formats', 'format_sheet']

# Internal Cell
import openpyxl
from openpyxl.styles import Alignment
from openpyxl.worksheet.table import Table, TableStyleInfo
import re
from pathlib import Path
import toolz.curried as tz
from fnmatch import fnmatch

from datetime import date
from dateutil.relativedelta import relativedelta as delta
from random import choice, randint, random
from typing import NewType, Dict, List, Iterable, Any, Optional, Union

names = (
    "Joe Nancy Harry Gus David Max Matt Nicole Tracy Stacy "
    "Kristi Kelly Mike James"
)
texts = (
    "Univeristy Franklin End_Game Tower New_York_Mets Last_Dance Movies Donald_Trump "
    "Nation_Captial COVID-19 Honesty Some_Really_Long_Text The_Lord_of_the_Rings"
)
flag = lambda: True if randint(1, 10) < 10 else False
text = lambda: choice([val.replace("_", " ") for val in texts.split()]) if flag() else None
name = lambda: choice(names.split())
dt = lambda: date.today() + delta(days=randint(-30, 30)) if flag() else None
amt = lambda: randint(150_000, 350_000) * random()

gen_data = lambda n=100: [
    {
        "Order Date": dt(),
        "Close Date": dt(),
        "Name": name(),
        "Dept": text(),
        "Balance": amt(),
        "Loan": amt(),
    }
    for row in range(n)
]

XLSXData = Iterable[Dict[Any,Any]]

XLSXSheetName = NewType("XLSXSheetName", str)
def to_xlsx(data: XLSXData, file: Path, sheet_name: XLSXSheetName,
    over_write: bool=True) -> Path:
    """Allows data to be appended to a .xlsx file as a new sheet."""
    if not over_write and file.exists():
        wb = openpyxl.load_workbook(file)
        ws = wb.create_sheet(sheet_name)
    else:
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.title = sheet_name
    for row_num, row in enumerate(data):
        if row_num == 0:
            ws.append([k for k in row])
        ws.append(v for v in row.values())
    wb.save(str(file))
    wb.close()
    return file

def make_test_file() -> Path:
    """Makes a test .xlsx file used for testing."""
    file = Path.cwd() / "tests/test.xlsx"
    sheet_name = "Test 1"
    over_write = True
    data = gen_data(100)
    return to_xlsx(data, file, sheet_name, over_write)

XLSXColumn = NewType("XLSXColumn", str)
XLSXFormatAttr = NewType("XLSXFormatAttr", str)
XLSXFormat = Dict[XLSXFormatAttr,Union[str, int, Alignment]]

def column_format(number_format: str="", width: int=15, vertical: str="top",
    horizontal: str="left", wrap_text: bool=True, **kwargs) -> XLSXFormat:
    """Returns a basic format allowing for text alignment, number_format,
    and column width to be set.
    """
    return {
        "number_format": number_format,
        "alignment": Alignment(vertical=vertical, horizontal=horizontal, wrap_text=wrap_text, **kwargs),
        "width": width,
    }

def default_formats(file: Path, sheet_name: XLSXSheetName,
    rows_to_review: int=100) -> Dict[XLSXColumn, XLSXFormat]:
    """Analyzes and existing sheet and provides default text alignment,
    column width, and number_format options based upon how the .xlsx
    file has stored the data.
    """
    wb = openpyxl.load_workbook(str(file), read_only=True)
    ws = wb[sheet_name]
    rows = ws.rows
    header = next(rows)
    # 1st row contains column headers
    names = [cell.value for cell in header]
    letters = [cell.column_letter for cell in header]
    return tz.pipe(
        # Loops over rows in sheet
        rows,
        tz.take(rows_to_review),
        # For each row, create a dict usng names has keys
        tz.map(lambda row: dict(zip(names, row))),
        # Get the .xlsx data_type for each cell
        tz.map(tz.valmap(tz.flip(getattr, "data_type"))),
        # Combine cells into a list per column
        tz.merge_with(list),
        # Count the cells for each data type in the column
        tz.valmap(tz.frequencies),
        lambda row: {
            k: (
                # If any cells have a date value
                column_format("YYYY-MM-DD", horizontal="right") if "d" in v else
                # If any cells have a string value
                column_format() if "s" in v else
                # Default data type in .xslx is "n" for number
                column_format("#,##0.00", horizontal="right")
            )
            for k, v in row.items()
        },
    )

UnixWildcard = NewType("UnixWildCard", str)
Pattern = Dict[UnixWildcard, XLSXFormat]

def column_formats(file: Path, sheet_name: XLSXSheetName, patterns=None):
    """Returns a column format for each column in the sheet using either
    provided formats or default formats.

    Provided formats for the 'patterns' argument use a Unix shell-style
    wildcard as the key and the value is an XLSXFormat. The XLSXFormat is
    typically created by using the column_format function.

    Unix shell-style pattern
    ╒═══════════╤══════════════════════════════════╕
    │ Pattern   │ Meaning                          │
    ╞═══════════╪══════════════════════════════════╡
    │ *         │ matches everything               │
    ├───────────┼──────────────────────────────────┤
    │ ?         │ matches any single character     │
    ├───────────┼──────────────────────────────────┤
    │ [seq]     │ matches any character in seq     │
    ├───────────┼──────────────────────────────────┤
    │ [!seq]    │ matches any character not in seq │
    ╘═══════════╧══════════════════════════════════╛
    """
    defaults = default_formats(file, sheet_name)
    if patterns is None:
        return defaults

    results = {}
    for col in defaults:
        for pat, fmt in patterns.items():
            if fnmatch(col, pat) and col not in results:
                results[col] = fmt
        if col not in results:
            results[col] = defaults[col]
    return results

XLSXCellReference = NewType("XLSXCellReference", str)
XLSXTableName = NewType("XLSXTableName", str)

def format_sheet(file: Path, sheet_name: XLSXSheetName,
    format_patterns: Optional[Pattern]=None,
    table_name: Optional[XLSXTableName]=None,
    freeze_panes: XLSXCellReference="A2") -> Path:
    """Formats a sheet and optionally adds the sheet as a
    Worksheet table with a default table format.

    Provided formats for the 'patterns' argument use a Unix shell-style
    wildcard as the key and the value is an XLSXFormat. The XLSXFormat is
    typically created by using the column_format function.

    Unix shell-style pattern
    ╒═══════════╤══════════════════════════════════╕
    │ Pattern   │ Meaning                          │
    ╞═══════════╪══════════════════════════════════╡
    │ *         │ matches everything               │
    ├───────────┼──────────────────────────────────┤
    │ ?         │ matches any single character     │
    ├───────────┼──────────────────────────────────┤
    │ [seq]     │ matches any character in seq     │
    ├───────────┼──────────────────────────────────┤
    │ [!seq]    │ matches any character not in seq │
    ╘═══════════╧══════════════════════════════════╛
    """
    # Load file
    wb = openpyxl.load_workbook(file)
    ws = wb[sheet_name]

    # Get column names and calc formats
    rows = ws.rows
    column_names = [cell.value for cell in next(rows)]
    formats = column_formats(file, sheet_name, format_patterns)

    # Loop by column and format column
    for col in ws.iter_cols():
        header = col[0]
        name = header.value
        letter = header.column_letter

        # Set width first so that text wrap works
        ws.column_dimensions[letter].width = formats[name]["width"]

        # Add alignment and number format
        for cell in col:
            cell.alignment = formats[name]["alignment"]
            cell.number_format = formats[name]["number_format"]

    # Freeze panes
    ws.freeze_panes = ws[freeze_panes]

    # Add table
    table_name = table_name or sheet_name
    table_name = table_name.replace(" ", "_")
    table = Table(displayName=table_name, ref=ws.dimensions)
    style = TableStyleInfo(name="TableStyleLight1", showFirstColumn=False, showColumnStripes=False, showRowStripes=True)
    table.tableStyleInfo = style
    ws.add_table(table)

    # Save file
    wb.save(str(file))
    wb.close()

    return file


from typing import List, Dict, Any, Iterator, Union, Callable, Sequence
import toolz.curried as tz
from decorator import decorator
from datetime import datetime, date
import re
from jedi_toolz.str_funcs import *

__all__ = (
    "to_table handle_data pretty_names"
).split()

RecordValue = Union[str, bool, int, float, date,
    datetime, None]
Record = Dict[str, RecordValue]
Table = Iterator[Record]


def has_all_attrs(obj: Any, attrs: str) -> bool:
    """Checks an object for all of the provided attrs."""
    assert isinstance(attrs, str)
    return all([hasattr(obj, attr)
        for attr in attrs.split()])

def has_any_attrs(obj: Any, attrs: str) -> bool:
    """Checks an object for any of the provided attrs."""
    assert isinstance(attrs, str)
    return any([hasattr(obj, attr)
        for attr in attrs.split()])

def is_pandas(data: Any) -> bool:
    """Returns True if data is a pandas.DataFrame."""
    assert data is not None
    attrs = (
        "values columns transpose head to_dict "
        "from_records"
    )
    return has_all_attrs(data, attrs)

def is_ipython() -> bool:
    """Returns True if the module is running inside an ipython kernal."""
    return hasattr(__builtins__,'__IPYTHON__')

def is_record(data: Any) -> bool:
    assert data is not None
    if not isinstance(data, dict):
        return False
    if not all([isinstance(k, str) for k in data]):
        return False
    else:
        return True

def is_table(data: Any) -> bool:
    assert data is not None

    if is_pandas(data): return False
    if isinstance(data, dict): return False

    attrs = "__iter__ __next__"
    if not has_any_attrs(data, attrs): return False

    first_row, _ = tz.peek(data)
    return is_record(first_row)

def to_table(data: Any) -> Table:
    assert data is not None

    if is_table(data):
        return data

    elif is_record(data):
        return [
            {"column": k, "value": v}
            for k, v in data.items()
        ]

    elif isinstance(data, dict):
        return [
            {"column": str(k), "value":v}
            for k, v in data.items()
        ]

    elif is_pandas(data):
        records = data.to_dict("records")
        return [
            {str(k): v for k, v in row.items()}
            for row in records
        ]

    else:
        raise ValueError(
            "data does not appear to be an Iterator of dicts, a "
            "dict, or a DataFrame."
        )

@decorator
def handle_data(func, *args, **kwargs):
    arg1, *other_args = args
    data = to_table(arg1)
    updated_args = data, *other_args
    return func(*updated_args, **kwargs)

@tz.curry
def pretty_names(data: Any, *funcs) -> Any:
    """Returns the Table with pretty names.
    uses the pretty_str function to modify names.
    """
    defaults = (
        decamel,
        invalid_first_char,
        replace_non_word,
        conseq_char,
        trim,
        lambda s: s.replace("_", " "),
    )
    if len(funcs) == 0:
        funcs = defaults

    if is_pandas(data):
        return data.rename(columns={
            col: fmt_str(col, *funcs)
            for col in data.columns
        })
    else:
        return [
            {fmt_str(k, *funcs): v for k, v in row.items()}
            for row in to_table(data)
        ]